The package {{ package }} failed to build, here is a possible explanation why.

Please know that the explanation was provided by AI and may be incorrect.
In this case, we are {{ certainty }}% certain of the response {{ emoji_face }}.

{{ explanation }}

<details>
<ul>
{% for snippet in snippets %}
<li>
<b>Line {{ snippet.line_number }}:</b> <code>{{ snippet.text }}</code>
{{ snippet.explanation }}
</li>
{% endfor %}
</ul>
</details>

<details>
  <summary>Logs</summary>
  <p>
    Log Detective analyzed the following logs files to provide an explanation:
  </p>

  <ul>
    <li><a href="{{ log_url }}">{{ log_url }}</a></li>
  </ul>

  <p>
    Additional logs are available from:
    <ul>
    <li><a href="{{ artifacts_url }}">artifacts.zip</a></li>
  </ul>
  </p>

  <p>
    Please know that these log files are automatically removed after some
    time, so you might need a backup.
  </p>
</details>

<details>
  <summary>Help</summary>
  <p>Don't hesitate to reach out.</p>

  <ul>
    <li><a href="https://github.com/fedora-copr/logdetective">Upstream</a></li>
    <li><a href="https://github.com/fedora-copr/logdetective/issues">Issue tracker</a></li>
    <li><a href="https://redhat.enterprise.slack.com/archives/C06DWNVKKDE">Slack</a></li>
    <li><a href="https://log-detective.com/documentation">Documentation</a></li>
  </ul>
</details>


---
This comment was created by [Log Detective][log-detective].

Was the provided feedback accurate and helpful? <br>Please vote with :thumbsup:
or :thumbsdown: to help us improve.<br>



[log-detective]: https://log-detective.com/
[contact]: https://github.com/fedora-copr
